"""Module for preservation of the legacy API, it will be removed in version 2.X of earthkit.transforms."""
from earthkit.transforms.climatology._aggregate import (
    anomaly,
    auto_anomaly,
    daily_max,
    daily_mean,
    daily_median,
    daily_min,
    daily_reduce,
    daily_std,
    max,
    mean,
    median,
    min,
    monthly_max,
    monthly_mean,
    monthly_median,
    monthly_min,
    monthly_reduce,
    monthly_std,
    percentiles,
    quantiles,
    reduce,
    relative_anomaly,
    std,
)

__all__ = [
    reduce,
    mean,
    median,
    min,
    max,
    std,
    daily_reduce,
    daily_mean,
    daily_median,
    daily_min,
    daily_max,
    daily_std,
    monthly_reduce,
    monthly_mean,
    monthly_median,
    monthly_min,
    monthly_max,
    monthly_std,
    quantiles,
    percentiles,
    anomaly,
    relative_anomaly,
    auto_anomaly,
]
