"""Module for preservation of the legacy API, it will be removed in version 2.X of earthkit.transforms."""
from earthkit.transforms.temporal._aggregate import (
    daily_max,
    daily_mean,
    daily_median,
    daily_min,
    daily_reduce,
    daily_std,
    daily_sum,
    max,
    mean,
    median,
    min,
    monthly_max,
    monthly_mean,
    monthly_median,
    monthly_min,
    monthly_reduce,
    monthly_std,
    monthly_sum,
    reduce,
    rolling_reduce,
    standardise_time,
    std,
    sum,
)

__all__ = [
    standardise_time,
    reduce,
    mean,
    median,
    min,
    max,
    std,
    sum,
    daily_reduce,
    daily_mean,
    daily_median,
    daily_max,
    daily_min,
    daily_std,
    daily_sum,
    monthly_reduce,
    monthly_mean,
    monthly_median,
    monthly_min,
    monthly_max,
    monthly_std,
    monthly_sum,
    rolling_reduce,
]
